(* ProCSV
 * Code generation: translate takes a semantically checked AST and
produces LLVM IR

LLVM tutorial: Make sure to read the OCaml version of the tutorial

http://llvm.org/docs/tutorial/index.html

Detailed documentation on the OCaml LLVM library:

http://llvm.moe/
http://llvm.moe/ocaml/

*)

module L = Llvm
module A = Ast
module StringMap = Map.Make(String)

let translate (globals, functions) =
  let context = L.global_context () in
  let the_module = L.create_module context "ProCSV"
  and i32_t  = L.i32_type  context
  and i8_t   = L.i8_type   context
  and i1_t   = L.i1_type   context
  and f_t    = L.double_type context
  and ptr_t = L.pointer_type (L.i8_type (context))
  and void_t = L.void_type context in

  let ltype_of_typ = function
      A.Int -> i32_t
    | A.Float -> f_t
    | A.Bool -> i1_t
    | A.String -> ptr_t
    | A.Void -> void_t in

  (* Declare each global variable; remember its value in a map *)
  let global_vars =
    let global_var m (t, n) =
      let init = L.const_int (ltype_of_typ t) 0
      in StringMap.add n (L.define_global n init the_module) m in
    List.fold_left global_var StringMap.empty globals in

  (* Declare printf(), which the print built-in function will call *)
  let printf_t = L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func = L.declare_function "printf" printf_t the_module in

  (* Declare the built-in printbig() function *)
  let printbig_t = L.function_type i32_t [| i32_t |] in
  let printbig_func = L.declare_function "printbig" printbig_t the_module in

  (* Declare the built-in read_csv() function *)
  let read_csv_t = L.function_type ptr_t [| ptr_t |] in
  let read_csv_func = L.declare_function "read_csv" read_csv_t the_module in
   
  (* Declare the built-in sim() function *) 
  let sim_t = L.function_type ptr_t [| ptr_t; ptr_t |] in 
  let sim_func = L.declare_function "sim" sim_t the_module in 

  (* Declare the built-in parse() function *) 
  let parse_t = L.function_type ptr_t [| ptr_t; ptr_t |] in 
  let parse_func = L.declare_function "parse" parse_t the_module in 

  (* Declare the built-in merge() function *) 
let merge_t = L.function_type ptr_t [| ptr_t; ptr_t; ptr_t |] in 
  let merge_func = L.declare_function "merge" merge_t the_module in 
       
  (* Declare the built-in find() function *) 
  let find_t = L.function_type ptr_t [| ptr_t; ptr_t |] in 
  let find_func = L.declare_function "find" find_t the_module in 

  (* Declare the built-in printbig() function 
  let print_string_t = L.function_type i8_t [| L.pointer_type i8_t |] in
  let print_string_func = L.declare_function "print_string" print_string_t the_module in*)

  (* Define each function (arguments and return type) so we can call it *)
  let function_decls =
    let function_decl m fdecl =
      let name = fdecl.A.fname
      and formal_types =
	     Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.A.formals)
      in let ftype = L.function_type (ltype_of_typ fdecl.A.typ) formal_types in
      StringMap.add name (L.define_function name ftype the_module, fdecl) m in
    List.fold_left function_decl StringMap.empty functions in
  
  (* Fill in the body of the given function *)
  let build_function_body fdecl =
    let (the_function, _) = StringMap.find fdecl.A.fname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in
    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder in
    let float_format_str  = L.build_global_stringptr "%f\n" "fmt" builder in
    (*let string_format_str  = L.build_global_stringptr "%s\n" "fmt" builder in*)
    
    (* Construct the function's "locals": formal arguments and locally
       declared variables.  Allocate each on the stack, initialize their
       value, if appropriate, and remember their values in the "locals" map *)
    let local_vars =
      let add_formal m (t, n) p = L.set_value_name n p;
	    let local = L.build_alloca (ltype_of_typ t) n builder in
	    ignore (L.build_store p local builder);
	    StringMap.add n local m in

      let add_local m (t, n) =
	    let local_var = L.build_alloca (ltype_of_typ t) n builder
	    in StringMap.add n local_var m in

      let formals = List.fold_left2 add_formal StringMap.empty fdecl.A.formals
          (Array.to_list (L.params the_function)) in
        List.fold_left add_local formals fdecl.A.locals in

    (* Return the value for a variable or formal argument *)
    let lookup n = try StringMap.find n local_vars
                   with Not_found -> StringMap.find n global_vars
    in
(* Construct code for an expression; return its value *) 
    let rec llvalue_expr_getter builder = function
          A.Id s -> lookup s
    |_ -> raise (Failure ("in llvalue_expr_getter but not a dotop!"))
and
  (* Return addr of lhs expr *)
  (*let addr_of_expr expr builder g_map l_map = match expr with
    A.Id(id) -> (lookup g_map l_map id)
   | _ -> raise (Failure("addr not found")) 

  in *)
    (* Construct code for an expression; return its value *)
   expr builder = function
	   A.Literal i -> L.const_int i32_t i
    | A.FloatLit f -> L.const_float f_t f
    | A.BoolLit b -> L.const_int i1_t (if b then 1 else 0)
    | A.StringLit s -> L.build_global_stringptr s "str" builder
    | A.Noexpr -> L.const_int i32_t 0
    | A.Id s -> L.build_load (lookup s) s builder
    | A.Binop (e1, op, e2) ->
	   let e1' = expr builder e1
	   and e2' = expr builder e2 in
        if (L.type_of e1' = f_t || L.type_of e2' = f_t) then
           (match op with
            A.Add     -> L.build_fadd
          | A.Sub     -> L.build_fsub
          | A.Mult    -> L.build_fmul
          | A.Div     -> L.build_fdiv
          | A.Equal   -> L.build_fcmp L.Fcmp.Oeq
          | A.Neq     -> L.build_fcmp L.Fcmp.One
          | A.Less    -> L.build_fcmp L.Fcmp.Olt
          | A.Leq     -> L.build_fcmp L.Fcmp.Ole
          | A.Greater -> L.build_fcmp L.Fcmp.Ogt
          | A.Geq     -> L.build_fcmp L.Fcmp.Oge
          | _ -> raise (Failure ("operator not supported for operand"))
          ) e1' e2' "tmp" builder
      else
	     (match op with
	     | A.Add     -> L.build_add
	     | A.Sub     -> L.build_sub
	     | A.Mult    -> L.build_mul
       | A.Div     -> L.build_sdiv
       | A.Mod     -> L.build_srem
	     | A.And     -> L.build_and
	     | A.Or      -> L.build_or
	     | A.Equal   -> L.build_icmp L.Icmp.Eq
	     | A.Neq     -> L.build_icmp L.Icmp.Ne
	     | A.Less    -> L.build_icmp L.Icmp.Slt
	     | A.Leq     -> L.build_icmp L.Icmp.Sle
	     | A.Greater -> L.build_icmp L.Icmp.Sgt
	     | A.Geq     -> L.build_icmp L.Icmp.Sge
	       ) e1' e2' "tmp" builder
  | A.Unop(op, e) ->
	 let e' = expr builder e in
	 (match op with
	  A.Neg     -> L.build_neg
    | A.Not     -> L.build_not) e' "tmp" builder
  
  | A.Pop(e, op) -> 
   let e' = expr builder e in
        (match op with
        | A.Inc -> ignore(expr builder (A.Assign(e, A.Binop(e, A.Add, A.Literal(1))))); e'                 
        | A.Dec -> ignore(expr builder (A.Assign(e, A.Binop(e, A.Sub, A.Literal(1))))); e') 

      
  (*|  A.Assign (e1, e2) -> let l_val = (addr_of_expr e1 builder g_map l_map) in
    let e2' = expr builder g_map l_map e2 in
    ignore (L.build_store e2' l_val builder); e2' *)
      | A.Assign (lhs, e2) -> let e2' = expr builder e2 in
      (match lhs with
      A.Id s ->ignore (L.build_store e2' (lookup s) builder); e2'
       |_ -> raise (Failure("nooo"))
          )
(*        |_ -> raise (Failure("can't match in assign"))
 *)

      | A.Call ("print", [e]) | A.Call ("printb", [e])->
	  L.build_call printf_func [| int_format_str ; (expr builder e) |]
	    "printf" builder
      | A.Call ("print_string", [e]) ->
      L.build_call printf_func [|(expr builder e) |] "print_string" builder
      | A.Call ("print_float", [e]) ->
      L.build_call printf_func [|float_format_str;(expr builder e)|] "print_float" builder
      | A.Call ("printbig", [e]) ->
	  L.build_call printbig_func [| (expr builder e) |] "printbig" builder
      | A.Call ("read_csv", [e]) ->
          L.build_call read_csv_func [| (expr builder e) |] "read_csv" builder
      (* | A.Call ("sim", args) ->
          L.build_call sim_func [|(expr builder e1) , (expr builder e2)|] "sim" builder *)
      | A.Call ("sim", act) ->
        let actuals = List.rev (List.map (expr builder) (List.rev act)) in
         L.build_call sim_func (Array.of_list actuals) "sim" builder   
      | A.Call ("parse", act) ->
        let actuals = List.rev (List.map (expr builder) (List.rev act)) in
         L.build_call parse_func (Array.of_list actuals) "parse" builder 
      | A.Call ("find", act) ->
        let actuals = List.rev (List.map (expr builder) (List.rev act)) in
         L.build_call find_func (Array.of_list actuals) "find" builder   
      | A.Call ("merge", act) ->
        let actuals = List.rev (List.map (expr builder) (List.rev act)) in
         L.build_call merge_func (Array.of_list actuals) "merge" builder   
      | A.Call (f, act) ->
         let (fdef, fdecl) = StringMap.find f function_decls in
	 let actuals = List.rev (List.map (expr builder) (List.rev act)) in
	 let result = (match fdecl.A.typ with A.Void -> ""
                                            | _ -> f ^ "_result") in
         L.build_call fdef (Array.of_list actuals) result builder
    in

    (* Invoke "f builder" if the current block doesn't already
       have a terminal (e.g., a branch). *)
    let add_terminal builder f =
      match L.block_terminator (L.insertion_block builder) with
	Some _ -> ()
      | None -> ignore (f builder) in
	
    (* Build the code for the given statement; return the builder for
       the statement's successor *)
    let rec stmt builder = function
	A.Block sl -> List.fold_left stmt builder sl
      | A.Expr e -> ignore (expr builder e); builder
      | A.Return e -> ignore (match fdecl.A.typ with
	  A.Void -> L.build_ret_void builder
	| _ -> L.build_ret (expr builder e) builder); builder
      | A.If (predicate, then_stmt, else_stmt) ->
         let bool_val = expr builder predicate in
	 let merge_bb = L.append_block context "merge" the_function in

	 let then_bb = L.append_block context "then" the_function in
	 add_terminal (stmt (L.builder_at_end context then_bb) then_stmt)
	   (L.build_br merge_bb);

	 let else_bb = L.append_block context "else" the_function in
	 add_terminal (stmt (L.builder_at_end context else_bb) else_stmt)
	   (L.build_br merge_bb);

	 ignore (L.build_cond_br bool_val then_bb else_bb builder);
	 L.builder_at_end context merge_bb
      | A.While (predicate, body) ->
	  let pred_bb = L.append_block context "while" the_function in
	  ignore (L.build_br pred_bb builder);

	  let body_bb = L.append_block context "while_body" the_function in
	  add_terminal (stmt (L.builder_at_end context body_bb) body)
	    (L.build_br pred_bb);

	  let pred_builder = L.builder_at_end context pred_bb in
	  let bool_val = expr pred_builder predicate in

	  let merge_bb = L.append_block context "merge" the_function in
	  ignore (L.build_cond_br bool_val body_bb merge_bb pred_builder);
	  L.builder_at_end context merge_bb

      | A.For (e1, e2, e3, body) -> stmt builder
	    ( A.Block [A.Expr e1 ; A.While (e2, A.Block [body ; A.Expr e3]) ] )
    in

    (* Build the code for each statement in the function *)
    let builder = stmt builder (A.Block fdecl.A.body) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder (match fdecl.A.typ with
        A.Void -> L.build_ret_void
      | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
  in

  List.iter build_function_body functions;
  the_module
